@@ -202,6 +202,10 @@ class ArrayImpl(basearray.Array):
   def sharding(self):
     return self._sharding
 
+  @property
+  def weak_type(self):
+    return self.aval.weak_type
+
   def __str__(self):
     return str(self._value)
 
@@ -277,23 +281,23 @@ class ArrayImpl(basearray.Array):
       return lax_numpy._rewriting_take(self, idx)
     else:
       # TODO(yashkatariya): Don't bounce to host and use `_rewriting_take` or
-      # the fast path (see PmapSharding branch above) after b/245667823 is
-      # fixed.
-      return self._value[idx]
+      # the fast path (see PmapSharding branch above) after after uneven
+      # partitioning support is added
+      return device_put(self._value[idx])
 
   def __iter__(self):
     if self.ndim == 0:
       raise TypeError("iteration over a 0-d array")  # same as numpy error
     else:
-      assert self.is_fully_replicated() or self.is_fully_addressable()
+      assert self.is_fully_replicated or self.is_fully_addressable
       if dispatch.is_single_device_sharding(self.sharding):
         return (sl for chunk in self._chunk_iter(100) for sl in chunk._unstack())  # type: ignore
       elif isinstance(self.sharding, PmapSharding):
         return (self[i] for i in range(self.shape[0]))  # type: ignore
       else:
         # TODO(yashkatariya): Don't bounce to host and use `_chunk_iter` path
-        # here after b/245667823 is fixed.
-        return (self._value[i] for i in range(self.shape[0]))
+        # here after uneven partitioning support is added.
+        return (device_put(self._value[i]) for i in range(self.shape[0]))
 
   def item(self):
     if dtypes.issubdtype(self.dtype, np.complexfloating):
@@ -307,6 +311,7 @@ class ArrayImpl(basearray.Array):
     else:
       raise TypeError(self.dtype)
 
+  @property
   def is_fully_replicated(self) -> bool:
     return self.shape == self._arrays[0].shape
 
@@ -317,7 +322,7 @@ class ArrayImpl(basearray.Array):
     else:
       dtype_str = f'dtype={self.dtype.name})'
 
-    if self.is_fully_addressable() or self.is_fully_replicated():
+    if self.is_fully_addressable or self.is_fully_replicated:
       line_width = np.get_printoptions()["linewidth"]
       s = np.array2string(self._value, prefix=prefix, suffix=',',
                           separator=', ', max_line_width=line_width)
@@ -329,8 +334,9 @@ class ArrayImpl(basearray.Array):
     else:
       return f"{prefix}{self.shape}, {dtype_str}"
 
+  @pxla.maybe_cached_property
   def is_fully_addressable(self) -> bool:
-    return self.sharding.is_fully_addressable()
+    return self.sharding.is_fully_addressable
 
   def __array__(self, dtype=None, context=None):
     return np.asarray(self._value, dtype=dtype)
@@ -371,7 +377,7 @@ class ArrayImpl(basearray.Array):
   # TODO(https://github.com/google/jax/issues/12380): Remove this when DA is
   # deleted.
   @property
-  def device_buffer(self) -> DeviceArray:
+  def device_buffer(self) -> ArrayImpl:
     self._check_if_deleted()
     if len(self._arrays) == 1:
       return _single_device_array_from_buf(self._arrays[0], self._committed)
@@ -381,7 +387,7 @@ class ArrayImpl(basearray.Array):
   # TODO(https://github.com/google/jax/issues/12380): Remove this when SDA is
   # deleted.
   @property
-  def device_buffers(self) -> Sequence[DeviceArray]:
+  def device_buffers(self) -> Sequence[ArrayImpl]:
     self._check_if_deleted()
     return [_single_device_array_from_buf(a, self._committed)
             for a in self._arrays]
@@ -446,12 +452,12 @@ class ArrayImpl(basearray.Array):
     self._check_if_deleted()
 
     if self._npy_value is None:
-      if self.is_fully_replicated():
+      if self.is_fully_replicated:
         self._npy_value = np.asarray(self._arrays[0])  # type: ignore
         self._npy_value.flags.writeable = False
         return cast(np.ndarray, self._npy_value)
 
-      if not self.is_fully_addressable():
+      if not self.is_fully_addressable:
         raise RuntimeError("Fetching value for `jax.Array` that spans "
                            "non-addressable devices is not possible. You can use "
                            "`jax.experimental.multihost_utils.process_allgather` "
@@ -520,9 +526,6 @@ mlir.register_constant_handler(ArrayImpl, _array_mlir_constant_handler)
 
 
 def _device_put_array(x, device: Optional[Device]):
-  # TODO(yashkatariya): Remove this restriction and the round trip via host
-  # once lowering to XLA goes through `lower_mesh_computation`.
-  assert x.is_fully_addressable()
   if dispatch.is_single_device_sharding(x.sharding):
     x = dispatch._copy_device_array_to_device(pxla._set_aval(x._arrays[0]), device)
     return (x,)
@@ -548,38 +551,25 @@ def _array_pmap_shard_arg(x, devices, indices, mode):
     return pxla._shard_sharded_device_array_slow_path(x, devices, indices, mode)
 
 
-def _array_rest_shard_arg(x, devices, indices, mode):
-  if not x._committed:
+def _array_rest_shard_arg(x: ArrayImpl, devices, indices, mode):
+  x_indices = x.sharding.addressable_devices_indices_map(x.shape).values()
+  if not x.is_fully_addressable:
+    if tuple(x_indices) == tuple(indices):
+      return x._arrays
+    else:
+      return NotImplementedError("Cannot reshard an input that is not fully "
+                                 "addressable")
+  else:
+    if tuple(x_indices) == tuple(indices):
+      return [buf if buf.device() == d else buf.copy_to_device(d)
+              for buf, d in safe_zip(x._arrays, devices)]
+    # Resharding starts here:
+    if isinstance(x.sharding, PmapSharding):
+      return pxla.device_put(x._value, devices, replicate=True)
     if dispatch.is_single_device_sharding(x.sharding):
-      # This condition is to break the recursion that happens when only
-      # `pxla._shard_device_array` is used since it has `_multi_slice` in the
-      # implementation which is jitted. Eventually it calls back here and the
-      # recursion happens.
-      x_indices = tuple(x.sharding.addressable_devices_indices_map(x.shape).values())
-      if x_indices == indices:
-        return [buf if buf.device() == d else buf.copy_to_device(d)
-                for buf, d in safe_zip(x._arrays, devices)]
       return pxla._shard_device_array(x, devices, indices, mode)
     else:
-      raise NotImplementedError('Resharding uncommitted arrays sharded over '
-                                'multiple devices is not supported.')
-  # TODO(yashkatariya): Remove the special case here and don't move to another
-  # device if its already committed. There is a TODO in dispatch.py already
-  # for this.
-  if dispatch.is_single_device_sharding(x.sharding):
-    return [buf if buf.device() == d else buf.copy_to_device(d)
-            for buf, d in safe_zip(x._arrays, devices)]
-  # If PmapSharding exists, then do a round trip via host. This will happen
-  # if the input Array containing PmapSharding takes the jit path
-  # i.e. `apply_primitive` or `xla_callable_uncached`. `jit(pmap)` is the most
-  # common case where this will happen.
-  # TODO(yashkatariya): Remove the special case here and don't move to another
-  # device if its already committed. There is a TODO in dispatch.py already
-  # for this.
-  elif isinstance(x.sharding, PmapSharding):
-    return pxla.device_put(x._value, devices, replicate=True)
-  else:
-    return x._arrays
+      return pxla._shard_sharded_device_array_slow_path(x, devices, indices, mode)
 
 
 def _array_shard_arg(x, devices, indices, mode):
