@@ -1,4 +1,4 @@
-# Copyright 2021 Google LLC
+# Copyright 2021 The JAX Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -78,6 +78,7 @@ _zero_preserving_unary_primitives = [
   lax.atan_p,
   lax.atanh_p,
   lax.bessel_i1e_p,
+  lax.copy_p,
   lax.expm1_p,
   lax.log1p_p,
   lax.neg_p,
@@ -187,6 +188,9 @@ class SparsifyValue(NamedTuple):
   def is_sparse(self):
     return self.indices_ref is not None
 
+  def is_dense(self):
+    return self.indices_ref is None
+
 
 _is_bcoo = lambda arg: isinstance(arg, BCOO)
 _is_spvalue = lambda arg: isinstance(arg, SparsifyValue)
@@ -453,23 +457,32 @@ def sparsify(f, use_tracer=False):
 #------------------------------------------------------------------------------
 # Sparse rules for various primitives
 
+def _ensure_unique_indices(spenv, spvalue):
+  """Return an spvalue representation with deduplicated indices."""
+  if spvalue.is_dense() or spvalue.unique_indices:
+    return spvalue
+  arr = spvalues_to_arrays(spenv, spvalue)
+  arr = arr.sum_duplicates(nse=arr.nse)
+  return arrays_to_spvalues(spenv, arr)
+
 def _zero_preserving_unary_op(prim):
   def func(spenv, *spvalues, **kwargs):
     assert len(spvalues) == 1
-    buf = spenv.data(spvalues[0])
+    # Since unary operations don't commute with addition, we need to ensure
+    # that indices are unique before applying the operator elementwise.
+    spvalue = _ensure_unique_indices(spenv, spvalues[0])
+    buf = spenv.data(spvalue)
     buf_out = prim.bind(buf, **kwargs)
     if spvalues[0].is_sparse():
-      out_spvalue = spenv.sparse(spvalues[0].shape, buf_out,
-                                 indices_ref=spvalues[0].indices_ref,
-                                 indices_sorted=spvalues[0].indices_sorted,
-                                 unique_indices=spvalues[0].unique_indices)
+      out_spvalue = spenv.sparse(spvalue.shape, buf_out,
+                                 indices_ref=spvalue.indices_ref,
+                                 indices_sorted=spvalue.indices_sorted,
+                                 unique_indices=spvalue.unique_indices)
     else:
       out_spvalue = spenv.dense(buf)
     return (out_spvalue,)
   return func
 
-# TODO(jakevdp): some of these will give incorrect results when there are duplicated indices.
-#                how should we handle this?
 for _prim in _zero_preserving_unary_primitives:
   sparse_rules[_prim] = _zero_preserving_unary_op(_prim)
 
@@ -746,6 +759,20 @@ def _todense_sparse_rule(spenv, spvalue, *, tree):
 
 sparse_rules[sparse.todense_p] = _todense_sparse_rule
 
+def _slice_sparse_rule(spenv, *operands, **params):
+  args = spvalues_to_arrays(spenv, operands)
+  out = sparse.bcoo_slice(*args, **params)
+  return arrays_to_spvalues(spenv, [out])
+
+sparse_rules[lax.slice_p] = _slice_sparse_rule
+
+def _dynamic_slice_sparse_rule(spenv, *operands, **params):
+  args = spvalues_to_arrays(spenv, operands)
+  out = sparse.bcoo_dynamic_slice(args[0], args[1:], **params)
+  return arrays_to_spvalues(spenv, [out])
+
+sparse_rules[lax.dynamic_slice_p] = _dynamic_slice_sparse_rule
+
 
 #------------------------------------------------------------------------------
 # BCOO methods derived from sparsify
@@ -762,6 +789,25 @@ def _reshape(self, *args, **kwargs):
 def _sparse_rewriting_take(arr, idx, indices_are_sorted=False, unique_indices=False,
                            mode=None, fill_value=None):
   # mirrors lax_numpy._rewriting_take.
+
+  # Handle some special cases, falling back if error messages might differ.
+  if (arr.ndim > 0 and isinstance(idx, (int, np.integer)) and
+      not isinstance(idx, (bool, np.bool_)) and isinstance(arr.shape[0], int)):
+    if 0 <= idx < arr.shape[0]:
+      return sparsify(lambda arr: lax.index_in_dim(arr, idx, keepdims=False))(arr)
+  if (arr.ndim > 0 and isinstance(arr.shape[0], int) and
+      isinstance(idx, slice) and
+      (type(idx.start) is int or idx.start is None) and
+      (type(idx.stop)  is int or idx.stop is  None) and
+      (type(idx.step)  is int or idx.step is  None)):
+    n = arr.shape[0]
+    start = idx.start if idx.start is not None else 0
+    stop  = idx.stop  if idx.stop  is not None else n
+    step  = idx.step  if idx.step  is not None else 1
+    if (0 <= start < n and 0 <= stop <= n and 0 < step and
+        (start, stop, step) != (0, n, 1)):
+      return sparsify(lambda arr: lax.slice_in_dim(arr, start, stop, step))(arr)
+
   treedef, static_idx, dynamic_idx = lax_numpy._split_index_for_jit(idx, arr.shape)
   result = sparsify(
       lambda arr, idx: lax_numpy._gather(arr, treedef, static_idx, idx, indices_are_sorted,
