@@ -1,4 +1,4 @@
-# Copyright 2018 Google LLC
+# Copyright 2018 The JAX Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -13,8 +13,8 @@
 # limitations under the License.
 
 from functools import partial
-from typing import (Any, Callable, Dict, Set, Optional, Tuple, Union, Iterable,
-                    Type, Sequence)
+from typing import (Any, Callable, Dict, Hashable, Iterable, Optional, Sequence,
+                    Set, Tuple, Type, Union)
 
 import numpy as np
 
@@ -36,15 +36,18 @@ from jax.interpreters import partial_eval as pe
 map = safe_map
 
 def _update_annotation(
-    f: lu.WrappedFun,
-    orig_type: Optional[Tuple[Tuple[core.AbstractValue, bool], ...]],
-    axis_size: int, axis_name: core.AxisName, in_dims: Sequence[Optional[int]]
-  ) -> lu.WrappedFun:
-  if orig_type is None:
-    return f
-  batched_in_type = [(core.unmapped_aval(axis_size, axis_name, dim, aval), keep)
-                     for dim, (aval, keep) in zip(in_dims, orig_type)]
-  return lu.annotate(f, tuple(batched_in_type))
+    f: lu.WrappedFun, orig_type: Optional[core.InputType],
+    axis_size: core.AxisSize, axis_name: core.AxisName,
+    in_dims: Sequence[Optional[int]]) -> lu.WrappedFun:
+  if orig_type is None: return f
+  if isinstance(axis_size, core.Tracer):
+    in_type_ = [(core.unmapped_aval(core.DBIdx(0), axis_name, dim, aval), keep)
+                for dim, (aval, keep) in zip(in_dims, orig_type)]
+    in_type = [(axis_size.aval, False), *in_type_]
+  else:
+    in_type = [(core.unmapped_aval(axis_size, axis_name, dim, aval), keep)
+               for dim, (aval, keep) in zip(in_dims, orig_type)]
+  return lu.annotate(f, tuple(in_type))
 
 ### vmappable typeclass
 
@@ -158,9 +161,11 @@ class BatchTracer(Tracer):
     return [('val', self.val), ('batch_dim', self.batch_dim)]
 
 class BatchTrace(Trace):
-  def __init__(self, *args, axis_name):
+
+  def __init__(self, *args, axis_name, spmd_axis_name = None):
     super().__init__(*args)
     self.axis_name = axis_name
+    self.spmd_axis_name = spmd_axis_name
 
   def pure(self, val):
     return BatchTracer(self, val, not_mapped, source_info_util.current())
@@ -174,6 +179,10 @@ class BatchTrace(Trace):
   def get_primitive_batcher(self, primitive, frame):
     if primitive in primitive_batchers:
       return primitive_batchers[primitive]
+    elif self.spmd_axis_name is not None and primitive in spmd_axis_primitive_batchers:
+      return partial(spmd_axis_primitive_batchers[primitive],
+                     self.spmd_axis_name, frame.size, frame.name,
+                     frame.main_trace.trace_type)
     elif primitive in axis_primitive_batchers:
       return self.get_axis_primitive_batcher(primitive, frame)
     msg = "Batching rule for '{}' not implemented"
@@ -187,7 +196,8 @@ class BatchTrace(Trace):
     if self.axis_name is core.no_axis_name:
       # If axis name is `no_axis_name` we can't find it via `core.axis_name` so
       # we reconstruct it from the information we have available
-      axis_size, = {x.shape[d] for x, d in zip(vals, dims) if d is not not_mapped}
+      axis_size, = core.dedup_referents(x.shape[d] for x, d in zip(vals, dims)
+                                        if d is not not_mapped)
       return core.AxisEnvFrame(self.axis_name, axis_size, self.main)
     return core.axis_frame(self.axis_name)
 
@@ -223,8 +233,9 @@ class BatchTrace(Trace):
       return call_primitive.bind(f, *vals, **params)
     else:
       f_, dims_out = batch_subtrace(f, self.main, dims)
-      ax_size, = {x.shape[d] for x, d in zip(vals, dims) if d is not not_mapped}
-      f_ = _update_annotation(f_, f.in_type, ax_size, self.axis_name, dims)
+      axis_size, = core.dedup_referents(x.shape[d] for x, d in zip(vals, dims)
+                                        if d is not not_mapped)
+      f_ = _update_annotation(f_, f.in_type, axis_size, self.axis_name, dims)
       vals_out = call_primitive.bind(f_, *vals, **params)
       src = source_info_util.current()
       return [BatchTracer(self, v, d, src) for v, d in zip(vals_out, dims_out())]
@@ -330,7 +341,7 @@ class BatchTrace(Trace):
     fun, out_dims1 = batch_subtrace(fun, self.main, in_dims)
     fwd, out_dims2 = batch_subtrace(fwd, self.main, in_dims)
     bwd = batch_custom_vjp_bwd(bwd, self.axis_name, axis_size,
-                               out_dims2, in_dims, self.main.trace_type)
+                               out_dims2, in_dims, self.main.trace_type, self.spmd_axis_name)
     out_vals = prim.bind(fun, fwd, bwd, *in_vals, out_trees=out_trees)
     fst, out_dims = lu.merge_linear_aux(out_dims1, out_dims2)
     if not fst:
@@ -362,7 +373,7 @@ class BatchTrace(Trace):
       return map(partial(BatchTracer, trace), vals, primal_dims, primal_srcs)
     def bwd_transform(bwd):
       return batch_custom_vjp_bwd(bwd, axis_name, axis_size, dims, (None,),
-                                  trace_type)
+                                  trace_type, self.spmd_axis_name)
     return vals, todo, bwd_transform
 
 def _main_trace_for_axis_names(main_trace: core.MainTrace,
@@ -377,14 +388,17 @@ def _main_trace_for_axis_names(main_trace: core.MainTrace,
 
 def batch(fun: lu.WrappedFun, axis_name: core.AxisName, axis_size,
           in_dims, out_dim_dests, main_type: Type[BatchTrace] = BatchTrace,
-          ) -> lu.WrappedFun:
+          spmd_axis_name: Optional[Hashable] = None) -> lu.WrappedFun:
   # we split up _batch_inner and _batch_outer for the leak checker
   f = _batch_inner(fun, axis_size, out_dim_dests)
-  return _batch_outer(f, axis_name, axis_size, in_dims, main_type)
+  return _batch_outer(f, axis_name, axis_size, in_dims, main_type,
+                      spmd_axis_name)
 
 @lu.transformation
-def _batch_outer(axis_name, axis_size, in_dims, main_type, *in_vals):
-  with core.new_main(main_type, axis_name=axis_name) as main:
+def _batch_outer(axis_name, axis_size, in_dims, main_type, spmd_axis_name,
+                 *in_vals):
+  with core.new_main(
+      main_type, axis_name=axis_name, spmd_axis_name=spmd_axis_name) as main:
     with core.extend_axis_env(axis_name, axis_size, main):
       with source_info_util.transform_name_stack('vmap'):
         outs = yield (main, in_dims, *in_vals), {}
@@ -553,9 +567,9 @@ def batch_custom_jvp_subtrace(main, in_dims, *in_vals):
                      out_tangent_bds, out_dims, out_tangents)
   yield out_primals + out_tangents, out_dims * 2
 
-def batch_custom_vjp_bwd(bwd, axis_name, axis_size, in_dims, out_dim_dests, main_type):
+def batch_custom_vjp_bwd(bwd, axis_name, axis_size, in_dims, out_dim_dests, main_type, spmd_axis_name):
   bwd, out_dims_thunk = batch_subtrace(bwd)
-  bwd_ = _batch_outer(bwd, axis_name, axis_size, in_dims, main_type)
+  bwd_ = _batch_outer(bwd, axis_name, axis_size, in_dims, main_type, spmd_axis_name)
   return _match_axes_and_sum(bwd_, axis_size, axis_name, out_dims_thunk, out_dim_dests)
 
 @lu.transformation
@@ -589,6 +603,7 @@ def _matchaxis_symbolic_zeros(axis_name, sz, name, src, dst, x, sum_match=False)
 BatchingRule = Callable[..., Tuple[Any, Union[int, Tuple[int, ...]]]]
 primitive_batchers : Dict[core.Primitive, BatchingRule] = {}
 axis_primitive_batchers: Dict[core.Primitive, Callable] = {}
+spmd_axis_primitive_batchers: Dict[core.Primitive, Callable] = {}
 
 def defvectorized(prim):
   primitive_batchers[prim] = partial(vectorized_batcher, prim)
@@ -610,12 +625,14 @@ def broadcast_batcher(prim, args, dims, **params):
       either an int indicating the batch dimension, or else `not_mapped`
       indicating no batching.
   """
-  shapes = {(x.shape, d) for x, d in zip(args, dims) if np.ndim(x)}
-  if len(shapes) == 1:
+  assert len(args) > 1
+  shape, dim = next((x.shape, d) for x, d in zip(args, dims)
+                    if d is not not_mapped)
+  if all(core.symbolic_equal_shape(shape, x.shape) and d == dim
+         for x, d in zip(args, dims) if np.ndim(x)):
     # if there's only agreeing batch dims and scalars, just call the primitive
-    d = next(d for d in dims if d is not not_mapped)
     out = prim.bind(*args, **params)
-    return (out, (d,) * len(out)) if prim.multiple_results else (out, d)
+    return (out, (dim,) * len(out)) if prim.multiple_results else (out, dim)
   else:
     # We pass size of 1 here because (1) at least one argument has a real batch
     # dimension and (2) all unmapped axes can have a singleton axis inserted and
